@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 
 # This Python program scans an IMAP Inbox and runs every
-# entry against SpamAssassin.  For any entries that match,
+# entry against SpamAssassin. For any entries that match,
 # the message is copied to another folder, and the original
 # marked or deleted.
 
@@ -11,7 +11,7 @@
 # You may use isbg under any OSI approved open source license
 # such as those listed at http://opensource.org/licenses/alphabetical
 
-version="0.98"
+version="0.99"
 
 from subprocess import Popen, PIPE
 
@@ -23,6 +23,8 @@ import getpass
 import getopt
 import string
 import socket
+import time
+import atexit
 
 try:
   from hashlib import md5
@@ -31,12 +33,12 @@ except ImportError:
 
 # You can specify your imap password using a command line option (--imappassword).
 # This however is a really bad idea since any user on the system can run
-# ps and see the command line arguments.  If you really must do it non-interactively
+# ps and see the command line arguments. If you really must do it non-interactively
 # then set the password here.
 
 imapuser=getpass.getuser()
 imaphost='localhost'
-imapport=0  # autodetect - 143 for standard connection, 993 for imaps
+imapport=0 # autodetect - 143 for standard connection, 993 for imaps
 usessl=0
 imappassword=None
 imapinbox="INBOX"
@@ -44,11 +46,18 @@ spaminbox="INBOX.spam"
 teachonly=0
 learnspambox=None
 learnhambox=None
+movehamto=None
 learnthendestroy=0
+interactive=sys.stdin.isatty()
 thresholdsize=120000 # messages larger than this aren't considered
 pastuidsfile=None
-passwordfilename=None  # where the password is stored if requested
-savepw=0               # save the password
+lockfile=None
+lockfilename=None
+ignorelockfile=0
+lockfilegraceminutes=240 
+spamc=False
+passwordfilename=None # where the password is stored if requested
+savepw=0              # save the password
 alreadylearnt="Message was already un/learned"
 
 # satest is what command is used test if the message is spam
@@ -61,6 +70,8 @@ spamflagscmd="+FLAGS.SILENT"
 spamflags="("
 # include the spamassassin report in the message placed in spaminbox
 increport=1
+# delete messages with a score higher then this
+deletehigherthen=0
 # expunge before quiting causing all messages marked for deletion
 # to actually be deleted
 expunge=0
@@ -78,18 +89,21 @@ exitcodes=0
 exitcodeok=0         # all went well
 exitcodenewmsgs=1    # there were new messages - none of them spam
                      #                         (if exitcodes is on)
-exitcodenewspam=2    # they were all spams    ( if exitcodes is on)
+exitcodenewspam=2    # they were all spams     (if exitcodes is on)
 exitcodenewmsgspam=3 # there were new messages and new spam
                      #                         (if exitcodes is on)
 exitcodeflags=10     # there were errors in the command line arguments
 exitcodeimap=11      # there was an IMAP level error
 exitcodespamc=12     # there was error when communicating between spamc and spamd
+                     #
+exitcodetty=20       # there was an error because we're not in an interactive tty
+exitcodelocked=30    # there's certainly another isbg running
 
 # IMAP implementation detail
 # Courier IMAP ignores uid fetches where more than a certain number are listed
 # so we break them down into smaller groups of this size
 uidfetchbatchsize=25
-# password saving stuff.  A vague level of obfuscation
+# password saving stuff. A vague level of obfuscation
 passwordhashlen=256 # should be a multiple of 16
 passwordhash=None
 
@@ -100,38 +114,44 @@ def usage(ec):
         sslmsg="""
   --ssl                 Make an SSL connection to the IMAP server"""
     sys.stderr.write("""isbg: IMAP Spam begone %s
-All options are optional
+
+All options are optional (\o/), default are between brackets
 
   --imaphost hostname   IMAP server name [%s]%s
   --imapuser username   Who you login as [%s]
   --imapinbox mbox      Name of your inbox folder [%s]
   --spaminbox mbox      Name of your spam folder [%s]
-  --teachonly           Don't search spam, just learn from folders [%s]
+  --teachonly           Don't search spam, just learn from folders
   --learnspambox mbox   Name of your learn spam folder [%s]
   --learnhambox mbox    Name of your learn ham folder [%s]
-  --learnthendestroy    Mark learnt messages for deletion [%s]
+  --movehamto mbox      Move ham to folder [%s]
+  --learnthendestroy    Mark learnt messages for deletion
   --maxsize numbytes    Messages larger than this will be ignored as they are
                         unlikely to be spam [%d]
   --noreport            Don't include the SpamAssassin report in the message
                         copied to your spam folder
   --flag                The spams will be flagged in your inbox
   --delete              The spams will be marked for deletion from your inbox
+  --deletehigherthen #  Delete any spam with a score higher then #
   --expunge             Cause marked for deletion messages to also be deleted
                         (only useful if --delete is specified)
   --verbose             Show IMAP stuff happening
   --spamc               Use spamc instead of standalone SpamAssassin binary
   --savepw              Store the password to be used in future runs
+  --noninteractive      Prevent interactive requests
+  --ignorelockfile      Don't stop is lock file is present
   --nostats             Don't print stats
   --exitcodes           Use different exitcodes (see doc)
+
 (Your inbox will remain untouched unless you specify --flag or --delete)
-  
-See http://wiki.github.com/ook/isbg for more details\n""" % (version, imaphost, sslmsg, imapuser, imapinbox, spaminbox, teachonly, learnspambox, learnhambox, learnthendestroy, thresholdsize))
+
+See http://redmine.ookook.fr/projects/isbg/wiki for more details\n""" % (version, imaphost, sslmsg, imapuser, imapinbox, spaminbox, learnspambox, learnhambox, movehamto, thresholdsize))
     sys.exit(ec)
 
-def errorexit(msg):
+def errorexit(msg, exitcode=exitcodeflags):
     sys.stderr.write(msg)
     sys.stderr.write("\nUse --help to see valid options and arguments\n")
-    sys.exit(exitcodeflags)
+    sys.exit(exitcode)
 
 def addspamflag(flag):
     global spamflags
@@ -162,12 +182,14 @@ def dehexof(x):
 
 # argument processing
 longopts=[ "imaphost=", "imapuser=", "imapinbox=", "spaminbox=",
-       "maxsize=", "noreport", "flag", "delete", "expunge", "verbose",
-       "trackfile=", "spamc", "ssl", "savepw", "nostats", "exitcodes",
-       "learnhambox=", "learnspambox=", "teachonly", "learnthendestroy",
+       "maxsize=", "noreport", "flag", "delete", "deletehigherthen=",
+       "expunge", "verbose", "trackfile=", "spamc", "ssl", "savepw",
+       "nostats", "exitcodes", "learnhambox=", "movehamto=",
+       "learnspambox=", "teachonly", "learnthendestroy", "noninteractive",
+       "ignorelockfile",
        # options not mentioned in usage
        "imappassword=", "satest=", "sasave=", "spamflagscmd=", "spamflags=",
-       "help", "version", "imapport=", "passwordfilename=" 
+       "help", "version", "imapport=", "passwordfilename=", "lockfilegraceminutes="
        ]
 
 try:
@@ -186,15 +208,25 @@ for p in opts:
             errorexit("Unrecognized size - "+p[1])
         if thresholdsize<1:
             errorexit("Size "+`thresholdsize`+" is too small")
+    elif p[0]=="--deletehigherthen":
+        try:
+            deletehigherthen=float(p[1])
+        except:
+            errorexit("Unrecognized score - "+p[1])
+        if deletehigherthen<1:
+            errorexit("Score "+`deletehigherthen`+" is too small")
     elif p[0]=="--imapport":
         imapport=int(p[1])
     elif p[0]=="--noreport":
         increport=0
+    elif p[0]=="--noninteractive":
+        interactive=0
     elif p[0]=="--flag":
         addspamflag("\\Flagged")
     elif p[0]=="--delete":
         addspamflag("\\Deleted")
     elif p[0]=="--spamc":
+        spamc=True
         satest=["spamc", "-c"]
         sasave=["spamc"]
     elif p[0]=="--expunge":
@@ -220,6 +252,12 @@ for p in opts:
         sys.exit(0)
     elif p[0]=="--trackfile":
         pastuidsfile=p[1]
+    elif p[0]=="--lockfilename":
+        lockfilename=p[1]
+    elif p[0]=="--ignorelockfile":
+        ignorelockfile=1
+    elif p[0]=="--lockfilegraceminutes":
+        lockfilegraceminutes = int(p[1])
     else:
         locals()[p[0][2:]]=p[1]
 
@@ -241,6 +279,15 @@ if pastuidsfile is None:
     res=hexof(m.digest())
     pastuidsfile=pastuidsfile+res
 
+if lockfilename is None:
+    lockfilename=os.path.expanduser("~"+os.sep+".isbg-lock")
+
+# Delete lock file
+def removelock():
+  os.remove(lockfilename)
+
+atexit.register(removelock)
+
 # Password stuff
 def getpw(data,hash):
     res=""
@@ -281,9 +328,24 @@ if passwordhash is None:
         passwordhash=passwordhash+m.digest()
 
 if verbose:
+    print "Lock file is", lockfilename
     print "Trackfile is", pastuidsfile
     print "SpamFlags are", spamflags
     print "Password file is", passwordfilename
+ 
+# Acquirelockfilename or exit
+if ignorelockfile:
+  if verbose:
+    print "Lock file is ignored. Continue."
+else:
+  if os.path.exists(lockfilename) and (os.path.getmtime(lockfilename) + (lockfilegraceminutes * 60) > time.time()):
+    if verbose:
+      print "\nLock file is present. Guessing isbg is already running. Exit."
+    exit(exitcodelocked)
+  else:
+    lockfile = open(lockfilename, 'w')
+    lockfile.write(`os.getpid()`)
+    lockfile.close()
 
 # Figure out the password
 if imappassword is None:
@@ -296,6 +358,8 @@ if imappassword is None:
         
     # do we have to prompt?
     if imappassword is None:
+        if not interactive:
+          errorexit("You need to specify your imap password and save it with the --savepw switch", exitcodeok)
         imappassword=getpass.getpass("IMAP password for %s@%s: " % (imapuser, imaphost))
 
     # Should we save it?
@@ -308,68 +372,10 @@ if imappassword is None:
         f.write(hexof(setpw(imappassword, passwordhash)))
         f.close()
 
-# pastuids keeps track of which uids we have already seen, so
-# that we don't analyze them multiple times.  We store its
-# contents between sessions by saving into a file as Python
-# code (makes loading it here real easy since we just source
-# the file)
-pastuids=[]
-try:
-    execfile(pastuidsfile)
-except:
-    pass
-# remember what pastuids looked like so that we can compare at the end
-origpastuids=pastuids[:]
-
-# This function gets the list of uids corresponding
-# to a message range
-gure=re.compile(r"[0-9]+ \(UID ([0-9]+)\)")
-def getuids(imap, low, high):
-    # return empty list for empty mailbox
-    if high<low: return []
-    # request message range
-    range=`low`+":"+`high`
-    res=imap.fetch(range, "UID")
-    assertok(res, 'fetch', range, 'UID')
-    res2=[]
-    for i in res[1]:
-        mo=gure.match(i)
-        if mo is None:
-            if verbose: print "getuids Eh?", i
-        else:
-            res2.append(mo.group(1))
-    return res2
-
-# This function gets the size of each message in the provided
-# list
-gsre=re.compile(r"[0-9]+ \(UID ([0-9]+) RFC822.SIZE ([0-9]+)\)")
-def getsizes(imap, msgs):
-    res2=[]
-
-    # Python really needs do - while
-    while 1:
-        if len(msgs)==0: break
-        if len(msgs)>uidfetchbatchsize:
-            msgmore=msgs[uidfetchbatchsize:]
-            msgs=msgs[:uidfetchbatchsize]
-        else:
-            msgmore=[]
-        msgs=string.join(msgs, ',')
-        res=imap.uid("FETCH", msgs, "(UID RFC822.SIZE)")
-        assertok(res, "uid fetch", msgs, "(UID RFC822.SIZE)")
-        for i in res[1]:
-            mo=gsre.match(i)
-            if mo is None:
-                if verbose: print "getsize Eh?", i
-            else:
-                res2.append((mo.group(2), mo.group(1)))
-        msgs=msgmore
-    return res2
-
 # Retrieve the entire message
 def getmessage(uid, append_to=None):
     res = imap.uid("FETCH", uid, "(RFC822)")
-    body=res[1][0][1]
+    assertok(res, 'uid fetch', uid, '(RFC822)')
     if res[0]!="OK":
         assertok(res, 'uid fetch', uid, '(RFC822)')
         try:
@@ -380,6 +386,8 @@ def getmessage(uid, append_to=None):
                 print "The message was probably deleted while we are running"
             if append_to:
                 append_to.append(uid)
+    else:
+      body=res[1][0][1]
     return body
 
 # This function makes sure that each lines ends in <CR><LF>
@@ -397,52 +405,11 @@ def assertok(res,*args):
     if verbose:
         print `args`, "=", res
     if res[0]!="OK":
-        sys.stderr.write("\n%s returned %s - aborting\n" % (`args`,  res ))
-        sys.exit(exitcodeimap)
-
-# This class implements imap over SSL.
-class IMAP4S(imaplib.IMAP4):
-    def __init__(self, host='', port=993): imaplib.IMAP4.__init__(self, host, port)
-
-    def open(self, host, port):
-        self.baresock=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        self.baresock.connect( (self.host, self.port) )
-        self.ssl=socket.ssl( self.baresock, None, None )
-
-    def read(self, size):
-        res=""
-        while len(res)<size:
-            res=res+self.ssl.read(size-len(res))
-        return res
-
-    def readline(self):
-        # We can only do one character of lookahead, so this is done character by character
-        res=""
-        last=0
-        while last!="\n":
-            last=self.ssl.read(1)
-            res=res+last
-        return res
-
-    def send(self, data):
-        while len(data):
-            l=self.ssl.write(data)
-            if l==len(data): break
-            data=data[l:]
-
-    def shutdown(self):
-        del self.ssl
-        self.baresock.close()
-    
-    def socket(self):
-        """Do not send or receive any data on the returned socket otherwise you
-        will break the ssl connection.  Only set socket options and that sort
-        of thing"""
-        return self.baresock
+        errorexit("\n%s returned %s - aborting\n" % (`args`, res ), exitcodeimap)
 
 # Main code starts here
 if usessl:
-    imap=IMAP4S(imaphost, imapport)
+    imap=imaplib.IMAP4_SSL(imaphost, imapport)
 else:
     imap=imaplib.IMAP4(imaphost,imapport)
 
@@ -450,6 +417,64 @@ else:
 res=imap.login(imapuser, imappassword)
 assertok(res, "login",imapuser, 'xxxxxxxx')
 
+# Spamassassion training
+if learnspambox:
+  if verbose: print "Teach SPAM to SA from:", learnspambox
+  res=imap.select(learnspambox, 0)
+  assertok(res, 'select', learnspambox)
+  s_tolearn = int(res[1][0])
+  s_learnt = 0
+  typ, uids = imap.uid("SEARCH", None, "ALL")
+  uids = uids[0].split()
+  for u in uids:
+      body = getmessage(u)
+      p=Popen(["spamc", "--learntype=spam"], stdin = PIPE, stdout = PIPE, close_fds = True)
+      try:
+        out = p.communicate(body)[0]
+      except:
+        continue
+      code = p.returncode
+      if code == 69 or code == 74:
+        errorexit("spamd is missconfigured (use --allow-tell)")
+      p.stdin.close()
+      if not out.strip() == alreadylearnt: s_learnt += 1
+      if verbose: print u, out
+      if learnthendestroy:
+        res = imap.uid("STORE", u, spamflagscmd, "(\\Deleted)")
+        assertok(res, "uid store", u, spamflagscmd, "(\\Deleted)")
+  if expunge:
+    imap.expunge()
+
+if learnhambox:
+  if verbose: print "Teach HAM to SA from:", learnhambox
+  res=imap.select(learnhambox, 0)
+  assertok(res, 'select', learnhambox)
+  h_tolearn = int(res[1][0])
+  h_learnt = 0
+  typ, uids = imap.uid("SEARCH", None, "ALL")
+  uids = uids[0].split()
+  for u in uids:
+      body = getmessage(u)
+      p=Popen(["spamc", "--learntype=ham"], stdin = PIPE, stdout = PIPE, close_fds = True)
+      try:
+        out = p.communicate(body)[0]
+      except:
+        continue
+      code = p.returncode
+      if code == 69 or code == 74:
+        errorexit("spamd is missconfigured (use --allow-tell)")
+      p.stdin.close()
+      if not out.strip() == alreadylearnt: h_learnt += 1
+      if verbose: print u, out
+      if movehamto:
+        res=imap.uid("COPY", u, movehamto)
+        assertok(res, "uid copy", u, movehamto)
+      if learnthendestroy or movehamto:
+        res = imap.uid("STORE", u, spamflagscmd, "(\\Deleted)")
+        assertok(res, "uid store", u, spamflagscmd, "(\\Deleted)")
+  if expunge or movehamto:
+    imap.expunge()
+
 uids=[]
 
 if not teachonly:
@@ -461,52 +486,52 @@ if not teachonly:
   res=imap.select(imapinbox, 1)
   assertok(res, 'select', imapinbox, 1)
 
-  # it returns number of messages in response
-  low=1
-  high=int(res[1][0])
+  # get the uids of all mails with a size less then the thresholdsize
+  typ, inboxuids = imap.uid("SEARCH", None, "SMALLER", thresholdsize)
+  inboxuids = inboxuids[0].split()
 
-  # get the corresponding UIDs
-  alluids=getuids(imap,low,high)
-
-  for i in alluids:
-      if i not in pastuids:
-          uids.append(i)
-
-  # for the uids we haven't seen before, get their sizes
-  # The code originally got both the UIDs and size at the
-  # same time.  This however took significantly longer as
-  # I assume it stat()ed and perhaps even opened every message,
-  # even the ones we had seen before
-  sizeduids=getsizes(imap, uids)
-  uids=[]
-  for i in sizeduids:
-      if int(i[0])>thresholdsize:
-          pastuids.append(i[1])
-          if verbose:
-              print i[1], "is", i[0], "bytes so it is being skipped"
-      else:
-          uids.append(i[1])
+  # pastuids keeps track of which uids we have already seen, so
+  # that we don't analyze them multiple times. We store its
+  # contents between sessions by saving into a file as Python
+  # code (makes loading it here real easy since we just source
+  # the file)
+  pastuids=[]
+  try:
+    execfile(pastuidsfile)
+  except:
+    pass
+  # remember what pastuids looked like so that we can compare at the end
+  origpastuids=pastuids[:]
+  
+  # filter away uids that was previously scanned
+  uids = [u for u in inboxuids if u not in pastuids]
 
 # Keep track of new spam uids
 spamlist=[]
 
+# Keep track of spam that is to be deleted
+spamdeletelist=[]
+
 # Main loop that iterates over each new uid we haven't seen before
 for u in uids:
-    # Double check
-    if u in pastuids: continue
     # Retrieve the entire message
     body = getmessage(u, pastuids)
 
     # Feed it to SpamAssassin in test mode
     p=Popen(satest, stdin=PIPE, stdout=PIPE, close_fds=True)
-    score = p.communicate(body)[0]
-    if score == "0/0":         
-      sys.stderr.write("spamc -> spamd error - aborting")
-      sys.exit(exitcodespamc)
+    try:
+      score = p.communicate(body)[0]
+      if not spamc:
+        m = re.search("score=(-?\d+(?:\.\d+)?) required=(\d+(?:\.\d+)?)", score)
+        score = m.group(1) + "/" + m.group(2) + "\n"
+    except:
+      continue
+    if score == "0/0\n":
+      errorexit("spamc -> spamd error - aborting", exitcodespamc)
 
     if verbose: print u, "score:", score
 
-    code  = p.returncode
+    code = p.returncode
     if code == 0:
         # Message is below threshold
         pastuids.append(u)
@@ -514,18 +539,25 @@ for u in uids:
         # Message is spam
         if verbose: print u, "is spam"
 
+        if deletehigherthen and float(score.split('/')[0]) > deletehigherthen:
+          spamdeletelist.append(u)
+          continue
+        
         # do we want to include the spam report
         if increport:
             # filter it through sa
             p = Popen(sasave, stdin = PIPE, stdout = PIPE, close_fds=True)
-            body = p.communicate(body)[0]
+            try:
+              body = p.communicate(body)[0]
+            except:
+              continue
             p.stdin.close()
             body=crnlify(body)
             res=imap.append(spaminbox, None, None, body)
             # The above will fail on some IMAP servers for various reasons.
             # we print out what happened and continue processing
             if res[0]!='OK':
-                print `["append", spaminbox, "{body}"]`, "failed for uid "+`u`+": "+`res`+".  Leaving original message alone."
+                print `["append", spaminbox, "{body}"]`, "failed for uid "+`u`+": "+`res`+". Leaving original message alone."
                 pastuids.append(u)
                 continue
         else:
@@ -535,69 +567,31 @@ for u in uids:
 
         spamlist.append(u)
 
+
+nummsg=len(uids)
+spamdeleted=len(spamdeletelist)
+numspam=len(spamlist)+spamdeleted
+
 # If we found any spams, now go and mark the original messages
-if len(spamlist):
+if numspam or spamdeleted:
     res=imap.select(imapinbox)
     assertok(res, 'select', imapinbox)
     # Only set message flags if there are any
-    if len(spamflags)>2:  
+    if len(spamflags)>2:
         for u in spamlist:
             res=imap.uid("STORE", u, spamflagscmd, spamflags)
             assertok(res, "uid store", u, spamflagscmd, spamflags)
             pastuids.append(u)
+    # Set deleted flag for spam with high score
+    for u in spamdeletelist:
+      res=imap.uid("STORE", u, spamflagscmd, "(\\Deleted)")
+      assertok(res, "uid store", u, spamflagscmd, "(\\Deleted)")
     if expunge:
       imap.expunge()
 
-
-nummsg=len(uids)
-numspam=len(spamlist)
-
-# Spamassassion training
-if learnspambox:
-  if verbose: print "Teach SPAM to SA from:", learnspambox
-  res=imap.select(learnspambox, 0)
-  assertok(res, 'select', learnspambox)
-  s_tolearn = int(res[1][0])
-  s_learnt  = 0
-  uids = getuids(imap, 1, s_tolearn)
-  for u in uids:
-      body = getmessage(u)
-      p=Popen(["spamc", "--learntype=spam"], stdin = PIPE, stdout = PIPE, close_fds = True)
-      out = p.communicate(body)[0]
-      p.stdin.close()
-      if not out.strip() == alreadylearnt: s_learnt += 1
-      if verbose: print u, out
-      if learnthendestroy:
-        res = imap.uid("STORE", u, spamflagscmd, "(\\Deleted)")
-        assertok(res, "uid store", u, spamflagscmd, "(\\Deleted)")
-
-if learnhambox:
-  if verbose: print "Teach HAM to SA from:", learnhambox
-  res=imap.select(learnhambox, 0)
-  assertok(res, 'select', learnhambox)
-  h_tolearn = int(res[1][0])
-  h_learnt  = 0
-  uids = getuids(imap, 1, h_tolearn)
-  for u in uids:
-      body = getmessage(u)
-      p=Popen(["spamc", "--learntype=ham"], stdin = PIPE, stdout = PIPE, close_fds = True)
-      out = p.communicate(body)[0]
-      p.stdin.close()
-      if not out.strip() == alreadylearnt: h_learnt += 1
-      if verbose: print u, out
-      if learnthendestroy:
-        res = imap.uid("STORE", u, spamflagscmd, "(\\Deleted)")
-        assertok(res, "uid store", u, spamflagscmd, "(\\Deleted)")
-  if expunge:
-    imap.expunge()
-
-
 if not teachonly:
   # Now tidy up lists of uids
-  newpastuids=[]
-  for i in pastuids:
-      if i in alluids and i not in newpastuids:
-          newpastuids.append(i)
+  newpastuids = list(set([u for u in pastuids if u in inboxuids]))
 
   # only write out pastuids if it has changed
   if newpastuids!=origpastuids:
@@ -619,11 +613,12 @@ del imap
 if stats:
   if learnspambox:
     print "%d/%d spams learnt" % (s_learnt, s_tolearn)
-  if learnhambox:  
+  if learnhambox:
     print "%d/%d hams learnt" % (h_learnt, h_tolearn)
   if not teachonly:
     print "%d spams found in %d messages" % (numspam, nummsg)
-    
+    print "%d/%d was automaticaly deleted" % (spamdeleted, numspam)
+
 if exitcodes and nummsg:
     res=0
     if numspam==0:
